from ._test import (
    logger,
    TestABC,
    SingleTest,
    MultiTestABC,
    ParallelMultiTest,
    SequentialMultiTest
)
from .enums import TestType, TestStatus
from .exceptions import TestFail, TestFailFatal
from . import enums, exceptions, tags, testinfo, testlib


__all__ = [
    'TestABC',
    'SingleTest',
    'MultiTestABC',
    'ParallelMultiTest',
    'SequentialMultiTest',
    'logger',
    'add_handler',
    'TestType',
    'TestStatus',
    'TestFail',
    'TestFailFatal',
    'enums',
    'exceptions',
    'tags',
    'testinfo',
    'testlib'
]



class TestFail(Exception):
    """Raised when a test fails on an expected condition."""
    pass


class TestFailFatal(Exception):
    """Raised when a test fails on an unexpected condition."""
    pass



from ..detector import *        # noqa: F401, F403
from ..element import *         # noqa: F401, F403
from ..pointmachine import *    # noqa: F401, F403
from ..route import *           # noqa: F401, F403
from ..zone import *            # noqa: F401, F403



import asyncio

from collections.abc import MutableMapping
from typing import (
    Any, Generic, Iterable, Iterator, Type, TypeVar, TypedDict, overload
)


_K = TypeVar("_K")
_V = TypeVar("_V")


class TestMemoryKeyMeta(TypedDict):
    """Metadata for a key in the test memory dictionary."""
    dtype: set[Type] | None
    """The data type(s) of the values associated with the key."""
    updatable: bool
    """If False, the value associated with the key cannot be overwritten."""
    lock: asyncio.Lock
    """An asyncio lock for the key to ensure data consistency."""


class TestMemoryKeyData(TypedDict, Generic[_V]):
    """Data for a key in the test memory dictionary."""
    value: _V | None
    """The value associated with the key. If None, the key has no value set."""
    history: list[_V] | None
    """A list of previous values for the key. If None, no history is stored."""
    meta: TestMemoryKeyMeta
    """The metadata dict for the key."""


class TestMemory(MutableMapping[_K, _V], Generic[_K, _V]):
    """
    A mapping-like structure for storing data collected during a test run.
    Provides basic functionality to create, read, write, and clear key-value
    pairs as well as additional features such as storing previous values for
    a key, managing data types, ensuring data consistency, and providing
    asynchronous access to the data.
    """

    def __init__(self) -> None:
        self.__memory: dict[_K, TestMemoryKeyData[_V]] = {}
        self.__memory_lock = asyncio.Lock()

    def __str__(self) -> str:
        return str(dict(self))

    def __repr__(self) -> str:
        return str(dict(self))

    def __getitem__(self, key: _K) -> _V:
        value = self.__memory[key]['value']
        if value is None:
            raise KeyError(f'Key "{key}" has no value assigned.')
        return value

    def __setitem__(self, key: _K, value: _V) -> None:
        if key in self.__memory:
            self._set_value(value, self._get_keydata(key), force_update=False)
        else:
            self.new_key(key, value=value)

    def __delitem__(self, key: _K) -> None:
        del self.__memory[key]

    def __iter__(self) -> Iterator[_K]:
        return iter(self.__memory)

    def __len__(self) -> int:
        return len(self.__memory)

    def _format_dtype(self, dtype: Any) -> set[Type] | None:
        """Format the data type(s) for a key value to a set of types or None.

        Args:
            dtype (Any): The data type(s) to be formatted.

        Raises:
            TypeError: If the data type is not a type or an iterable of types.

        Returns:
            set[Type] | None: The formatted data type(s) as a set of types\
                or None.
        """
        if dtype is None:
            return None
        if isinstance(dtype, Iterable) and not isinstance(dtype, str):
            dtype = set(t for t in dtype)
        else:
            dtype = set((dtype,))
        for t in dtype:
            if not isinstance(t, type):
                raise TypeError(
                    f'Invalid data type "{t}". Expected a type, '
                    f'got an object instance of type "{type(t).__name__}".'
                )
        return dtype

    def _validate_value(
        self, value: _V | None, keymeta: TestMemoryKeyMeta
    ) -> None:
        """Ensure that the value is of the correct data type for the key.

        Args:
            value (_V | None): The value to be validated.
            keymeta (TestMemoryKeyMeta): The metadata for the key associated\
                with the value.

        Raises:
            TypeError: If the value is not of the correct data type.
        """
        if value is None:
            return
        allowed_types = keymeta['dtype']
        if not allowed_types:
            return
        if any(isinstance(value, t) for t in allowed_types):
            return
        allowed_types_str = ', '.join(f'"{t.__name__}"' for t in allowed_types)
        raise TypeError(
            f'Invalid value "{value}" of type "{type(value).__name__}". '
            f'Expected a value of type(s) {allowed_types_str}.'
            )

    def _get_keydata(self, key: _K) -> TestMemoryKeyData[_V]:
        """Retrieve the data associated with the specified key from the memory
        dictionary.

        Args:
            key (_K): The key for which to retrieve the data.

        Raises:
            KeyError: If the key is not present in the memory dictionary.

        Returns:
            TestMemoryKeyData[_V]: The data associated with the key.
        """
        if key not in self.__memory:
            raise KeyError(f'Key "{key}" is not present in the test memory.')
        return self.__memory[key]

    def _set_value(
        self, value: _V, keydata: TestMemoryKeyData, force_update: bool
    ) -> None:
        """Set the value for the specified key in the memory dictionary.

        Args:
            value (_V): The value to be set for the key.
            keydata (TestMemoryKeyData): The data associated with the key.
            force_update (bool): If True, allows updating a non-updatable key.

        Raises:
            KeyError: If the key is not updatable and the "force_update"\
                parameter is set to False.
        """
        self._validate_value(value, keydata['meta'])
        value_prev = keydata['value']
        if value_prev is not None:
            if not keydata['meta']['updatable'] and not force_update:
                raise KeyError(
                    f'Cannot set updatable key value to "{value}".'
                    )
            if keydata['history'] is not None:
                keydata['history'].append(value_prev)
        keydata['value'] = value

    def new_key(
        self,
        key: _K,
        value: _V | None = None,
        dtype: Type | Iterable[Type] | None = None,
        history: bool = False,
        updatable: bool = True,
        overwrite: bool = False
    ) -> None:
        """Create a new key in the memory dictionary with optional data type
        and storage settings.

        Args:
            key (_K): The key to be created.
            value (_V, optional): The value to be associated with the key.
            dtype (Type | Iterable[Type] | None, optional): The data type\
                of the values associated with the key. If None, the value\
                associated with the key may have any value type(s).\
                Defaults to None.
            history (bool, optional): If True, previous values for this key\
                will be stored when new values are written. Defaults to False.
            updatable (bool, optional): If False, the value associated with\
                the key cannot be overwritten accidentally. Defaults to False.
            overwrite (bool, optional): If True, the key will be overwritten\
                if it already exists in the memory dictionary.\
                Defaults to False.

        Raises:
            KeyError: If the key already exists in the memory dictionary and\
                the "overwrite" parameter is set to False.
        """
        if key in self.__memory and not overwrite:
            raise KeyError(f'Key "{key}" already exists in the test memory.')
        keymeta: TestMemoryKeyMeta = {
            'dtype': self._format_dtype(dtype),
            'updatable': updatable,
            'lock': asyncio.Lock()
            }
        if value is not None:
            self._validate_value(value, keymeta)
        self.__memory[key] = {
            'value': value,
            'history': [] if history else None,
            'meta': keymeta
            }

    @overload
    async def read(self, key: _K, depth: None) -> _V | None: ...

    @overload
    async def read(self, key: _K, depth: int) -> list[_V]: ...

    async def read(
        self, key: _K, depth: int | None = None
    ) -> _V | list[_V] | None:
        """Retrieve the value associated with the specified key from the memory
        dictionary.

        This method acts as a standard dictionary read operation, but is also
        asynchronous and should be used in an asynchronous context. The method
        can also retrieve a specified number of previous values for the key
        if the "depth" parameter is set to a positive integer and the key has
        a history of previous values.

        Args:
            key (_K): The key for which to retrieve the value.
            depth (int | None, optional): The number of previous values to\
                retrieve for the key. If None, only the current value\
                is returned. Defaults to None.

        Returns:
            _V | list[_V] | None: The value associated with the key.\
                If the "depth" parameter is set to a positive integer,\
                a list of the previous values for the key will be returned.

        Raises:
            KeyError: If the key is not present in the test memory.
        """
        async with self.__memory_lock:
            keydata = self._get_keydata(key)
        async with keydata['meta']['lock']:
            value = keydata['value']
            if value is None:
                return None
            if depth is None:
                return value
            history = keydata['history']
            if history is None:
                return [value]
            if depth < 1:
                return [value]
            return history[-depth:] + [value]

    async def write(
        self, key: _K, value: _V, force_update: bool = False
    ) -> None:
        """
        Write a value to the specified key in the memory dictionary.

        This method acts as a standard dictionary write operation, but is also
        asynchronous and should be used in an asynchronous context.

        Args:
            key (_K): The key to write the value to.
            value (_V): The value to be written to the key.
            force_update (bool, optional): If set to True,\
                allows updating a non-updatable key. Defaults to False.

        Raises:
            KeyError: if attempting to reassign a non-updatable key\
                without setting the "force_update" parameter to True.
        """
        async with self.__memory_lock:
            if key not in self.__memory:
                self.new_key(key, value)
                return
            else:
                keydata = self._get_keydata(key)
        async with keydata['meta']['lock']:
            self._set_value(value, keydata, force_update)



import enum

from functools import cached_property, wraps
from typing import Any, Callable, Iterable, Iterator, Literal, NamedTuple

from ..enums import TestStatus


class LogFilter(enum.StrEnum):
    """
    Enumeration that defines log filter options for tagged methods
    in test classes.

    Note that the log filter only governs the "outer" logging during test
    execution; internal log messages remain unaffected.

    WARNING: Currently, setting this for a tagged method does not change
    logging behavior.
    """
    NEVER = enum.auto()
    """Do not log method execution."""
    IF_TRUE = enum.auto()
    """Log method execution only if it returns `True`."""
    IF_FALSE = enum.auto()
    """Log method execution only if it returns `False`."""
    ALWAYS = enum.auto()
    """Always log method execution."""


class TagType(enum.StrEnum):
    """
    Enumeration that defines types of tags applicable to methods
    in test classes.
    """
    BEFORE_COMMAND = enum.auto()
    """Method that is executed before a test command."""
    AFTER_COMMAND = enum.auto()
    """Method that is executed after a test command."""
    FAIL = enum.auto()
    """Method that evaluates whether the test has failed."""
    WARNING = enum.auto()
    """Method that evaluates whether the test has produced a warning."""
    SUCCESS = enum.auto()
    """Method that evaluates whether the test has succeeded."""
    ON_STATUS_CHANGE = enum.auto()
    """Method that is executed when the test status changes."""
    COMMAND = enum.auto()
    """Method that represents a test command."""
    VALUE_GETTER = enum.auto()
    """Method that acquires the value of the test."""
    DESCRIPTION = enum.auto()
    """Method that provides a description of the test."""


class TagInfo(NamedTuple):
    """
    A names tuple that holds tag information for a method in a test class,
    including the type of tag, the target command, priority, logging filter,
    and an optional description.
    """
    type_: TagType
    """The type of the tag."""
    target: Any | None
    """The target of the tag (e.g., the command name)."""
    priority: int
    """The priority of the tagged method execution."""
    log: LogFilter
    """The log filter for the tagged method."""
    description: str | None
    """The description of the tagged method."""


class TagDecorator:
    """
    A decorator factory that provides a mechanism for dynamically attaching
    metadata, or "tags," to methods in test classes. These tags categorize
    methods based on their function and importance in the overall test
    execution process.

    New decorators can be created using the class method `new`. The produced
    decorator can be used with parentheses, as in `@decorator(...)`, or
    without them, as in `@decorator`.

    Attributes:
        TAG_ATTR (str): Attribute name used to attach tags to methods.\
            Stores tags in a list of `TagInfo` named tuples.
    """

    TAG_ATTR: str = "_tags"
    """
    Name of the attribute used to store a list of `TagInfo` objects
    attached to a method.
    """

    def __init__(self, tags: TagInfo | Iterable[TagInfo]) -> None:
        """Initialize the `TagDecorator` with a set of tags that can be
        attached to methods.

        Args:
            tags (TagInfo | Iterable[TagInfo]): A single `TagInfo` or \
                an iterable of `TagInfo` instances to be used as tags.
        """
        self.tags: list[TagInfo] = self.format_tags(tags)

    def __call__(self, func: Callable) -> Callable:
        """Makes `TagDecorator` instance callable, allowing it to be used
        directly as a decorator without parameters. This method is invoked when
        the decorator is applied directly to a function, as in `@decorator`.

        When applied, it attached predefined tags to the function.
        Execution of the function is not affected.

        Args:
            func (Callable): The method to decorate with tags.

        Returns:
            Callable: A wrapper that decorates the method with tags,
            adding specified tag metadata to the method.
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return self.attach_tags(wrapper)

    @classmethod
    def new(
        cls, func: Callable | None = None, *, tags: TagInfo | Iterable[TagInfo]
    ) -> Callable:
        """Creates a decorator with specified tags.

        This method acts as a factory for creating a new `TagDecorator`
        instance. The decorator can be used with or without parentheses.

        When called without a function argument (`func`), it returns
        a decorator configured with the specified tags. This returned decorator
        can then be applied to a function using parentheses, as in
        `@decorator(...)`. If `func` is provided, it applies the decorator
        immediately to that function.

        Args:
            func (Callable | None, optional): The method to decorate.
            tags (TagInfo | Iterable[TagInfo]): Tags to attach to the method.

        Returns:
            Callable: Either the decorated method or a decorator\
                configured with specific tags, depending on the usage.
        """
        decorator = cls(tags)
        if func is None:  # the decorator was called with parentheses
            return decorator
        else:
            return decorator(func)

    @classmethod
    def get_func_tags(cls, func: Callable) -> list[TagInfo] | None:
        """Get the tags attached to a method, if any.

        Args:
            func (Callable): The method to get the tags from.

        Returns:
            list[TagInfo] | None: The list of `TagInfo` named tuples\
                assigned to the method or `None` if no tags are assigned.
        """
        return getattr(func, cls.TAG_ATTR, None)

    @staticmethod
    def format_tags(tags: TagInfo | Iterable[TagInfo]) -> list[TagInfo]:
        """Validate and format the tags into a list of `TagInfo` named tuples.

        Args:
            tags (TagInfo | Iterable[TagInfo]): The tags to format.

        Raises:
            ValueError: If the tags are of an invalid type.

        Returns:
            list[TagInfo]: The list of `TagInfo` named tuples.
        """
        if isinstance(tags, TagInfo):
            return [tags]
        elif isinstance(tags, Iterable):
            return [tag for tag in tags]
        raise ValueError(f'Invalid tag type "{type(tags).__name__}".')

    def attach_tags(self, func: Callable) -> Callable:
        """Associate tags with a method, storing them within the method's
        attributes under the `TAG_ATTR` attribute name.

        Args:
            func (Callable): The method to attach the tags to.

        Returns:
            Callable: The method with the tags attached.
        """
        if not hasattr(func, self.TAG_ATTR):
            setattr(func, self.TAG_ATTR, [])
        assert isinstance((tags := getattr(func, self.TAG_ATTR)), list)
        for tag in self.tags:
            tags.append(tag)
        return func


def get_obj_name(obj: Any) -> str:
    """Get the name of the object via the `__name__` attribute. If the object
    does not have this attribute, the object is converted to a string.

    Args:
        obj (Any): The object to get the name of.

    Returns:
        str: The name of the object.
    """
    try:
        return obj.__name__
    except AttributeError:
        return str(obj)


class ObjectTagCollector:
    """
    A utility class for collecting and organizing methods tagged with metadata
    within a given object. This class helps to categorize methods by their tags
    and provides facilities for accessing and sorting these methods based on
    their relevance to specific operations, such as test execution flow.
    """
    def __init__(self, obj: object) -> None:
        """Initialize the `ObjectTagCollector`.

        The methods that have been tagged using the `TagDecorator` are
        collected and organized them based on the tags.

        Args:
            obj (object): The object to collect tagged methods from.
        """
        self._obj = obj
        self.tagged_methods = self.collect_tagged_methods(obj)

    def _get_method_tags(self, method_name: str) -> list[TagInfo]:
        """Get the tags attached to the specified method.

        Args:
            method_name (str): The name of the method to get the tags from.

        Returns:
            list[TagInfo]: The list of `TagInfo` named tuples assigned to the\
                method.
        """
        method = getattr(self._obj, method_name)
        return TagDecorator.get_func_tags(method) or []

    def _get_method_priority(
        self, method_name: str, matches: Callable[[TagInfo], bool]
    ) -> int:
        """Get the priority of the method based on the specified tag matching
        function. This function is used to evaluate which tag to get
        the priority from when multiple tags are attached to the method.

        Args:
            method_name (str): The name of the method to get the priority of.
            matches (Callable[[TagInfo], bool]): The function that matches\
                the tag to the specified criteria to choose which tag to get\
                the priority from. It should accept a `TagInfo` named tuple\
                and return a boolean value.

        Raises:
            TypeError: If no matching tag attached to the method is found.

        Returns:
            int: The priority of the method.
        """
        method_tags = self._get_method_tags(method_name)
        for tag in method_tags:
            if matches(tag):
                return tag.priority
        # should be unreachable
        raise TypeError(f'Method "{method_name}" was not mached.')

    def _sort_methods_by_priority(
        self, method_names: Iterable[str], matches: Callable[[TagInfo], bool]
    ) -> list[str]:
        """Sort the methods by priority and then alphabetically by name for
        ties.

        Args:
            method_names (Iterable[str]): The names of the methods to sort.
            matches (Callable[[TagInfo], bool]): The function that matches\
                the tag to the specified criteria.

        Returns:
            list[str]: The sorted list of method names.
        """
        # get priority value for each method
        method_names_priorities = [
            (name, self._get_method_priority(name, matches))
            for name in method_names
            ]
        # sort by priority first, and then alphabetically by name for ties
        method_names_priorities.sort(key=lambda t: (t[1], t[0]))
        return [name for name, _ in method_names_priorities]

    @staticmethod
    def list_obj_methods(
        obj: object, exclude_dunder_methods: bool = True
    ) -> Iterator[str]:
        """List the method names of the specified object, excluding properties
        and optionally excluding dunder methods.

        Args:
            obj (object): The object to list the methods of.
            exclude_dunder_methods (bool, optional): Whether to exclude\
                dunder methods from the list. Defaults to `True`.

        Yields:
            Iterator[str]: A generator that yields method names of the object.
        """
        for method_name in dir(obj):
            method = getattr(obj, method_name, None)
            if (
                callable(method) and
                not isinstance(method, (property, cached_property)) and
                not (exclude_dunder_methods and method_name.startswith("__"))
            ):
                yield method_name

    def collect_tagged_methods(self, obj: object) -> dict[TagType, list[str]]:
        """Collect tagged methods from the specified object.

        Args:
            obj (object):  The object from which to collect tagged methods.

        Returns:
            dict[TagType, list[str]]: The dictionary categorizing methods by\
                their tag types. Dictionary keys are `TagType` values, and\
                the corresponding values are lists of method names.
        """
        tagged: dict[TagType, set[str]] = {tag: set() for tag in TagType}
        # collect tagged methods
        for method_name in self.list_obj_methods(obj):
            method_tags = self._get_method_tags(method_name)
            for tag in method_tags:
                tagged[tag.type_].add(method_name)
        # sort tagged methods by priority
        tagged_sorted: dict[TagType, list[str]] = {}
        for tag_type, method_names in tagged.items():
            tagged_sorted[tag_type] = self._sort_methods_by_priority(
                method_names, lambda tag: tag.type_ == tag_type
                )
        return tagged_sorted

    def count_tags(self, tag_type: TagType) -> int:
        """Get the number of methods tagged with the specified tag type.

        Args:
            tag_type (TagType): The tag type to count.

        Returns:
            int: The number of methods tagged with the specified tag type.
        """
        return len(self.tagged_methods.get(tag_type, []))

    def get_tagged_methods(self, tag_type: TagType) -> list[str]:
        """Get the methods tagged with the specified tag type.

        Args:
            tag_type (TagType): The type of tag to filter methods by.

        Returns:
            list[str]: The list of method names tagged with\
                the specified tag type.
        """
        return self.tagged_methods.get(tag_type, [])

    def get_method(self, method_name: str) -> Callable:
        """Get the method callable object from the object by its name.

        Args:
            method_name (str): The name of the method to get.

        Returns:
            Callable: The method object with the specified name.
        """
        return getattr(self._obj, method_name)

    def get_method_priority(self, method_name: str, tag_type: TagType) -> int:
        """Get the priority of a method within the context of a specified tag
        type. This method iterates over tags attached to the method and matches
        them against the specified tag type to determine the relevant priority.

        Args:
            method_name (str): The name of the method to get the priority of.
            tag_type (TagType): The tag type to match.

        Returns:
            int: The priority of the method within the specified tag type.
        """
        return self._get_method_priority(
            method_name, lambda tag: tag.type_ == tag_type
            )

    @cached_property
    def status_change_actions(self) -> dict[TestStatus, list[str]]:
        """
        A dictionary of the possible test status values and corresponding
        methods that are triggered when the status changes to this value.
        The list of method names is sorted by priority first, and then
        alphabetically by name for ties.
        """
        actions: dict[TestStatus, set[str]] = {ts: set() for ts in TestStatus}
        # collect tagged methods that are to be executed on status change
        tag_type = TagType.ON_STATUS_CHANGE
        for method_name in self.tagged_methods.get(tag_type, []):
            method_tags = self._get_method_tags(method_name)
            for tag in method_tags:
                if tag.type_ == tag_type:
                    assert isinstance(tag.target, TestStatus)
                    actions[tag.target].add(method_name)
        # sort status change actions by priority
        actions_sorted: dict[TestStatus, list[str]] = {}
        for status, method_names in actions.items():
            actions_sorted[status] = self._sort_methods_by_priority(
                method_names, lambda tag: tag.target == status
                )
        return actions_sorted

    @cached_property
    def commands_with_conditions(self) -> dict[
        str | None,
        dict[Literal[TagType.BEFORE_COMMAND, TagType.AFTER_COMMAND], list[str]]
    ]:
        """
        A dictionary of test command methods and corresponding routines that
        are executed before and after the commands. Both the commands
        (the dictionary keys) and the routines (lists mapped to
        the "TagType.BEFORE_COMMAND" and "TagType.AFTER_COMMAND" keys within
        the inner dictionary) are sorted by priority first, and then
        alphabetically by name for ties.
        """
        commands: dict = {
            command: {TagType.BEFORE_COMMAND: [], TagType.AFTER_COMMAND: []}
            for command in self.tagged_methods.get(TagType.COMMAND, [])
            }
        if None not in commands:
            commands[None] = {
                TagType.BEFORE_COMMAND: [], TagType.AFTER_COMMAND: []
                }
        for tag_type in [TagType.BEFORE_COMMAND, TagType.AFTER_COMMAND]:
            for method_name in self.tagged_methods.get(tag_type, []):
                method_tags = self._get_method_tags(method_name)
                for tag in method_tags:
                    if tag.type_ == tag_type:
                        target = tag.target
                        if target is None or isinstance(target, str):
                            conditions = commands[target][tag_type]
                            conditions.append(method_name)
                        else:
                            target_type = type(target).__name__
                            raise TypeError(
                                f'Invalid target type "{target_type}".'
                                )
        return commands

    @cached_property
    def value_getter(self) -> str | None:
        """
        A method that implements the value acqusition logic. Only one method
        within a class can be tagged as a value getter! If more than one
        method is tagged as a value getter, a ValueError is raised. If no
        method is tagged as a value getter, `None` is returned.
        """
        getter = self.tagged_methods.get(TagType.VALUE_GETTER, [])
        if len(getter) == 0:
            return None
        if len(getter) == 1:
            return getter[0]
        getters = ", ".join(f'"{g}"' for g in getter)
        raise ValueError(
            f'Test cannot have more than 1 value getter '
            f'(found {len(getter)}: {getters}).'
            )


def command_to_list(obj: Any) -> list[str] | list[None]:
    """Convert an object to a list of bounded commands.

    Args:
        obj (Any): The object to convert.

    Raises:
        ValueError: If the object cannot be converted to a list of bounded\
            commands.

    Returns:
        list[str] | list[None]: A list of bounded commands.
    """
    if obj is None:
        return [None]
    if callable(obj):
        obj = get_obj_name(obj)
    if isinstance(obj, str):
        return [obj]
    if isinstance(obj, Iterable):
        objs = []
        for o in obj:
            if callable(o):
                objs.append(get_obj_name(o))
            else:
                objs.append(o)
        return objs
    raise ValueError(
        f'Object of type "{type(obj).__name__}" cannot be converted to '
        f'a list of bounded commands.'
        )


def test_status_to_list(obj: Any) -> list[TestStatus]:
    """Convert an object to a list of bounded test statuses.

    Args:
        obj (Any): The object to convert.

    Raises:
        ValueError: If the object cannot be converted to a list of bounded\
            test statuses.

    Returns:
        list[TestStatus]: A list of bounded test statuses.
    """
    if obj is None:
        return [s for s in TestStatus]
    if isinstance(obj, str) or isinstance(obj, TestStatus):
        return [TestStatus(obj)]
    if isinstance(obj, Iterable):
        objs = []
        for o in obj:
            if isinstance(o, str):
                objs.append(TestStatus(o))
            else:
                objs.append(o)
        return objs
    raise ValueError(
        f'Object of type "{type(obj).__name__}" cannot be converted to '
        f'a list of bounded test statuses.'
        )


def description(method: Callable | None = None) -> Callable:
    """A decorator that marks a method as a dynamic description generator
    within a test class. Although the decorator does not expect any arguments
    except for the method itself, it can be used directly (without parentheses)
    or with parentheses. The latter makes no difference in the decorator's
    behavior.

    The decorated method should expect no arguments and return a string.
    It may be asynchronous. Only one method in any test class should be
    decorated with `@description`.
    Note:
        Method decorated with `@description` do not influence the flow of test
        execution, but are used to dynamically generate descriptions for
        reporting or logging purposes.

    Example:
        ```python
        class MyTest(SingleTest):
            @description
            def my_description(self) -> str:
                return f'Test description for {self.name}'
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tag = TagInfo(TagType.DESCRIPTION, None, 0, LogFilter.NEVER, None)
    return TagDecorator.new(method, tags=tag)


def command(
    method: Callable | None = None,
    /,
    *,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """A decorator that marks a method as a test command within a test class.
    A command is a method that represents a specific action or operation that
    is executed before the test evaluation. The decorator may be used directly
    without parentheses or with parentheses to specify additional parameters.

    The command priority can be set using the `priority` parameter. The lower
    the priority value, the command is executed. Commands with the same
    priority are executed in the alphabetical order of the method names.

    The `log` parameter specifies the log filter value for the command. See
    the `LogFilter` enumeration for possible values. WARNING: Currently,
    setting this for a tagged method does not change logging behavior.

    Description of the command action purpose can be specified using
    the `description` parameter.

    The decorated method should expect no arguments. It may return a value of
    any type. The decorated method may be asynchronous. The number of methods
    in a test class that can be decorated with `@command` is not limited.

    Example:
        ```
        # In this example, the `increment_by_one` and `increment_by_two`
        # methods are marked as commands. The `priority` parameter determines
        # the order of execution: `increment_by_one` will be executed before
        # `increment_by_two`.

        class MyTest(SingleTest):
            @command(priority=1, description="Increment the value by one")
            def increment_by_one(self):
                self.value += 1

            @command(priority=2, description="Increment the value by two")
            def increment_by_two(self):
                self.value += 2
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        priority (int, optional): Priority of the command. Lower values\
            indicate higher priority. Commands with the same priority\
            are evaluated in alphabetical order of their method names.\
            Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for\
            the command. See `LogFilter` enumeration for possible values.
            Defaults to LogFilter.ALWAYS. Note: Current implementation does not
            alter logging behavior.
        description (str | None, optional): Description of the purpose of\
            the command action. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tag = TagInfo(TagType.COMMAND, None, priority, log, description)
    return TagDecorator.new(method, tags=tag)


def before_command(
    method: Callable | None = None,
    /,
    *,
    command: str | Iterable[str] | Callable | Iterable[Callable] | None = None,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """A decorator that marks a method as a prerequisite condition for one or
    more test commands. This condition must be met before the specified
    command(s) can be executed. The decorator may be used directly without
    parentheses or with parentheses to specify additional parameters.

    The target command can be specified in the `command` parameter as a string,
    a list of strings, a callable, or a list of callables. If more than one
    command is specified, the condition is applied to all of them. If no
    command is specified, the condition must be met before the first command
    is executed. If a test does not implement commands, the condition must be
    met before the test evaluation.

    The prerequisite condition priority can be set using the `priority`
    parameter. The lower the priority value, the earlier the condition is
    checked. Conditions with the same priority are checked in the alphabetical
    order of the method names.

    The `log` parameter specifies the log filter value for the condition. See
    the `LogFilter` enumeration for possible values. WARNING: Currently,
    setting this for a tagged method does not change logging behavior.

    Description of the condition purpose can be specified using
    the `description` parameter.

    The decorated method should expect no arguments and return a boolean value
    or `None`. In the latter case, the condition is aumatically considered met.
    The decorated method may be asynchronous. The number of methods in a test
    class that can be decorated with `@before_command` is not limited.

    Example:
        ```
        # In this example, the "MyTest" class implements a command "increment"
        # that adds 1 to the value. To ensure that the value is 0 before the
        # command is applied, the "is_value_zero" method is marked as a
        # precondition for the "increment" method. Now, the "increment" method
        # will not be executed until the "is_value_zero" method returns True.

        class MyTest(SingleTest):

            @before_command(
                command='increment',
                description="Check if the initial value is 0"
                )
            def is_value_zero(self) -> bool:
                return self.value == 0

            @command(description="Increment the value by 1")
            def increment(self):
                self.value += 1
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        command (str | Iterable[str] | Callable | Iterable[Callable] | None,\
        optional): Command(s) the condition applies to. If None, the condition\
            must be met before the first command is executed. Defaults to None.
        priority (int, optional): Priority of the condition. Lower values\
            indicate higher priority. Conditions with the same priority\
            are evaluated in alphabetical order of their method names.\
            Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for\
            the condition. See `LogFilter` enumeration for possible values.
            Defaults to LogFilter.ALWAYS. Note: Current implementation does not
            alter logging behavior.
        description (str | None, optional): Description of the purpose of\
            the prerequisite condition. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tags = [
        TagInfo(TagType.BEFORE_COMMAND, target, priority, log, description)
        for target in command_to_list(command)
        ]
    return TagDecorator.new(method, tags=tags)


def after_command(
    method: Callable | None = None,
    /,
    *,
    command: str | Iterable[str] | Callable | Iterable[Callable] | None = None,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """A decorator that marks a method as a post-condition for one or
    more test commands. This post-condition is executed after the specified
    command(s) have been executed. The decorator may be used directly without
    parentheses or with parentheses to specify additional parameters.

    The target command can be specified in the `command` parameter as a string,
    a list of strings, a callable, or a list of callables. If more than one
    command is specified, the post-condition is applied to all of them. If no
    command is specified, the post-condition is executed after the last command
    is executed. If a test does not implement commands, the post-condition is
    executed before the test evaluation.

    The post-condition's execution priority can be set using the `priority`
    parameter. The lower the priority value, the earlier the post-condition is
    executed after the command(s). Post-conditions with the same priority are
    executed in the alphabetical order of the method names.

    The `log` parameter specifies the log filter value for the post-condition.
    See the `LogFilter` enumeration for possible values. WARNING: Currently,
    setting this for a tagged method does not change logging behavior.

    Description of the post-condition's purpose can be specified using
    the `description` parameter.

    The decorated method should expect no arguments and return a boolean value
    or `None`. In the latter case, the post-condition is automatically
    considered met. The decorated method may be asynchronous. The number of
    methods in a test class that can be decorated with `@after_command` is not
    limited.

    Example:
        ```
        # In this example, the "MyTest" class implements a command "decrement"
        # that subtracts 1 from the value. To check if the value is
        # non-negative after the "decrement" command is applied,
        # the "is_non_negative" method is marked as a post-condition for
        # the "decrement" method. Now, the "is_non_negative" method will be
        # executed immediately after the "decrement" method is called.

        class MyTest(SingleTest):

            @command(description="Decrement the value by 1")
            def decrement(self):
                self.value -= 1

            @after_command(
                command='decrement',
                description="Check if the final value is non-negative"
                )
            def is_non_negative(self) -> bool:
                return self.value >= 0
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        command (str | Iterable[str] | Callable | Iterable[Callable] | None,
        optional): Command(s) that the post-condition applies to. If None,\
            the post-condition is executed after the last command is executed.\
            Defaults to None.
        priority (int, optional): Priority of the post-condition. Lower values\
            indicate higher priority. Post-conditions with the same priority\
            are executed in alphabetical order of their method names.\
            Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for the\
            post-condition. See `LogFilter` enumeration for possible values.\
            Defaults to LogFilter.ALWAYS. Note: Current implementation\
            does not alter logging behavior.
        description (str | None, optional): Description of the purpose of\
            the post-condition. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tags = [
        TagInfo(TagType.AFTER_COMMAND, target, priority, log, description)
        for target in command_to_list(command)
    ]
    return TagDecorator.new(method, tags=tags)


def value_getter(
    method: Callable | None = None,
    /,
    *,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None,
) -> Callable:
    """A decorator that marks a method as a value getter within a test class.
    The value getter is a method that acquires the value that is a subject of
    the test evaluation. The decorator may be used directly without parentheses
    or with parentheses to specify additional parameters.

    The `log` parameter specifies the log filter value for the post-condition.
    See the `LogFilter` enumeration for possible values. WARNING: Currently,
    setting this for a tagged method does not change logging behavior.

    Description of the post-condition's purpose can be specified using
    the `description` parameter.

    The decorated method should expect no arguments and return a value of any
    type. It may be asynchronous. Only one method in any test class should be
    decorated with `@value_getter`.

    Example:
        ```
        # In this example, the "MyTest" class defines a method
        # "get_current_value" that simply returns the sum of 1 and 1.
        # This method is executed within the test evaluation loop to acquire
        # the current value. The return value of the "get_current_value" method
        # is stored under the "value" attribute of the test instance.
        # The "is_value_two" method then checks if the value is equal to 2.

        class MyTest(SingleTest):

            @value_getter(description="Get the current value")
            def get_current_value(self) -> int:
                return 1 + 1

            @success(description="Check if the value is 2")
            def is_value_two(self) -> bool:
                return self.value == 2
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        log (LogFilter, optional): Specifies the logging behavior for the\
            value getter. See `LogFilter` enumeration for possible values.\
            Defaults to LogFilter.ALWAYS. Note: Current implementation\
            does not alter logging behavior.
        description (str | None, optional): Description of the retrieved\
            value. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tag = TagInfo(TagType.VALUE_GETTER, None, 0, log, description)
    return TagDecorator.new(method, tags=tag)


def fail(
    method: Callable | None = None,
    /,
    *,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """A decorator that marks a method as a fail condition within a test class.
    If at least one of such conditions is met, the test is considered failed.
    The decorator may be used directly without parentheses or with parentheses
    to specify additional parameters.

    The execution priority of fail conditions can be set using the `priority`
    parameter. The lower the priority value, the earlier the fail condition is
    checked. Fail conditions with the same priority are checked in the
    alphabetical order of the method names.

    The `log` parameter specifies the log filter value for the fail condition.
    See the `LogFilter` enumeration for possible values. WARNING: Currently,
    setting this for a tagged method does not change logging behavior.

    Description of the fail condition can be specified using
    the `description` parameter.

    The decorated method should expect no arguments and return a boolean value
    or `None`. If the method returns `True`, the fail condition is considered
    met. If the method returns `False` or `None`, the condition is considered
    not met. The decorated method may be asynchronous. The number of methods
    in a test class that can be decorated with `@fail` is not limited.

    Example:
        ```
        # In this example, the "MyTest" class defines a method
        # "is_value_negative", which checks if the value is negative.
        # If the method returns True, the test is considered failed.

        class MyTest(SingleTest):
            @fail(description="Check if the value is negative")
            def is_value_negative(self) -> bool:
                return self.value < 0
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        priority (int, optional): Priority of the fail condition. Lower values\
            indicate higher priority. Fail conditions with the same priority\
            are executed in alphabetical order of their method names.\
            Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for the\
            fail condition. See `LogFilter` enumeration for possible values.\
            Defaults to LogFilter.ALWAYS. Note: Current implementation\
            does not alter logging behavior.
        description (str | None, optional): Description of the fail condition.\
            Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tag = TagInfo(TagType.FAIL, None, priority, log, description)
    return TagDecorator.new(method, tags=tag)


def warning(
    method: Callable | None = None,
    /,
    *,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """A decorator that marks a method as a warning condition within a test
    class. If this condition is met, it triggers a warning but does not
    necessarily fail the test. The decorator may be used directly without
    parentheses or with parentheses to specify additional parameters.

    The execution priority of warning conditions can be set using
    the `priority` parameter. The lower the priority value, the earlier
    the warning condition is checked. Warning conditions with the same priority
    are checked in the alphabetical order of the method names.

    The `log` parameter specifies the log filter value for the warning
    condition. See the `LogFilter` enumeration for possible values.
    WARNING: Currently, setting this for a tagged method does not change
    logging behavior.

    A description of the warning condition can be specified using
    the `description` parameter.

    The decorated method should expect no arguments and return a boolean value
    or `None`. If the method returns `True`, the warning condition is
    considered met. If the method returns `False` or `None`, the condition is
    considered not met. The decorated method may be asynchronous.
    The number of methods in a test class that can be decorated with `@warning`
    is not limited.

    Example:
        ```
        # In this example, the "MyTest" class defines a method "is_value_zero",
        # which checks if the value is zero. If the method returns True,
        # a warning is triggered.

        class MyTest(SingleTest):
            @warning(description="Check if the value is zero")
            def is_value_zero(self) -> bool:
                return self.value == 0
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        priority (int, optional): Priority of the warning condition. Lower\
            values indicate higher priority. Warning conditions with the same\
            priority are executed in alphabetical order of their method names.\
            Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for\
            the warning condition. See `LogFilter` enumeration for possible\
            values. Defaults to LogFilter.ALWAYS. Note: Current implementation\
            does not alter logging behavior.
        description (str | None, optional): Description of the warning\
            condition. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when
            the decorator is applied directly, or the decorator itself
            when used with parentheses.
    """
    tag = TagInfo(TagType.WARNING, None, priority, log, description)
    return TagDecorator.new(method, tags=tag)


def success(
    method: Callable | None = None,
    /,
    *,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """A decorator that marks a method as a success condition within a test
    class. Every success condition must be met for the test to be considered
    successful. The decorator may be used directly without parentheses or
    with parentheses to specify additional parameters.

    The execution priority of success conditions can be set using
    the `priority` parameter. The lower the priority value, the earlier
    the success condition is checked. Success conditions with the same priority
    are checked in the alphabetical order of the method names.

    The `log` parameter specifies the log filter value for the success
    condition. See the `LogFilter` enumeration for possible values.
    WARNING: Currently, setting this for a tagged method does not change
    logging behavior.

    Description of the success condition can be specified using
    the `description` parameter.

    The decorated method should expect no arguments and return a boolean value
    or `None`. If the method returns `True` or `None`, the success condition is
    considered met. If the method returns `False`, the condition is considered
    not met. The decorated method may be asynchronous. The number of methods
    in a test class that can be decorated with `@success` is not limited.

    Example:
        ```
        # In this example, the "MyTest" class defines a method
        # "is_value_nonnegative", which checks if the value is non-negative.
        # If the method returns True, the test is considered successful.

        class MyTest(SingleTest):
            @success(description="Check if the value is non-negative")
            def is_value_nonnegative(self) -> bool:
                return self.value >= 0
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.
            Usually, this argument should not be provided explicitly.
            If None, the decorator prepares to be applied when the method
            is passed in a subsequent call.
        priority (int, optional): Priority of the success condition. Lower\
            values indicate higher priority. Success conditions with the same\
            priority are executed in alphabetical order of their method names.\
            Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for\
            the success condition. See `LogFilter` enumeration for possible\
            values. Defaults to LogFilter.ALWAYS. Note: Current implementation\
            does not alter logging behavior.
        description (str | None, optional): Description of the success\
            condition. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when
            the decorator is applied directly, or the decorator itself
            when used with parentheses.
    """
    tag = TagInfo(TagType.SUCCESS, None, priority, log, description)
    return TagDecorator.new(method, tags=tag)


def on_status_change(
    method: Callable | None = None,
    /,
    *,
    change_to: (
        str | Iterable[str] | TestStatus | Iterable[TestStatus] | None
        ) = None,
    priority: int = 999,
    log: LogFilter = LogFilter.ALWAYS,
    description: str | None = None
) -> Callable:
    """Decorator for marking a method as an action to be performed when a test
    status changes to a specific status value(s). The method decorated with
    this decorator will be executed when the test's status changes to the
    specified status value(s). The decorator may be used directly without
    parentheses or with parentheses to specify additional parameters.

    The triggering status value(s) can be specified in the `change_to`
    parameter as a string, a list of strings, a member of the `TestStatus`
    enumeration, or a list of `TestStatus` members. If more than one status
    value is specified, the action is executed for all of them. If no status
    value is specified, the action is executed when the test status changes
    to any value.

    Priority of the action execution can be set using the `priority` parameter.
    The lower the priority value, the earlier the action is triggered after
    the status change. Actions with the same priority are executed in
    the alphabetical order of the method names.

    The `log` parameter specifies the log filter value for the action.
    See the `LogFilter` enumeration for possible values. WARNING: Currently,
    setting this for a tagged method does not change logging behavior.

    Description of the action's purpose can be specified using
    the `description` parameter.

    The decorated method should expect no arguments. It may return value of
    any type. The decorated method may be asynchronous. The number of actions
    in a test class that can be triggered by a status change is not limited.
    Overall number of methods in a test class that can be decorated with
    `@on_status_change` is not limited, too.

    Example:
        ```
        # In this example, the "MyTest" class implements methods that react to
        # status changes. The "log_status_change" method logs every status
        # change at the INFO level. The "log_timeout" method logs a warning
        # when the status changes to `TestStatus.TIMEOUT`.

        class MyTest(SingleTest):

            @on_status_change(
                description="log status change"
                )
            def log_status_change(self) -> None:
                self.logger.info("status changed to %s", self.status)

            @on_status_change(
                change_to=TestStatus.TIMEOUT, description="log timeout"
                )
            def log_timeout(self) -> None:
                self.logger.warning("timeout")
        ```

    Args:
        method (Callable | None, optional): The method to be decorated.\
            Usually, this argument should not be provided explicitly.\
            If None, the decorator prepares to be applied when the method\
            is passed in a subsequent call.
        change_to (str | Iterable[str] | TestStatus | Iterable[TestStatus] |
        None, optional): Status value(s) that trigger the action. If None,\
            the action is executed when the test status changes to any value.\
        priority (int, optional): Priority of the action within all actions\
            triggered by the same status change event. Lower values indicate\
            higher priority. Actions with the same priority are executed in\
            alphabetical order of their method names. Defaults to 999.
        log (LogFilter, optional): Specifies the logging behavior for the\
            action execution. See `LogFilter` enumeration for possible values.\
            Defaults to LogFilter.ALWAYS. Note: Current implementation\
            does not alter logging behavior.
        description (str | None, optional): Description of the purpose of\
            the action. Defaults to None.

    Returns:
        Callable: A decorated method with the attached description tag when\
            the decorator is applied directly, or the decorator itself\
            when used with parentheses.
    """
    tags = [
        TagInfo(TagType.ON_STATUS_CHANGE, target, priority, log, description)
        for target in test_status_to_list(change_to)
        ]
    return TagDecorator.new(method, tags=tags)



from __future__ import annotations

from datetime import datetime
from typing import Self
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

from .._test import TestABC, MultiTestABC, TestType, TestStatus


class TestInfoABC(BaseModel):
    """
    Test info model ABC.
    """
    model_config = ConfigDict(
        alias_generator=lambda s: to_camel(s.rstrip('_')),
        populate_by_name=True,
        validate_assignment=True,
        validate_default=True
        )


class TestInfo(TestInfoABC):
    """
    Test info model.
    """
    id_: int
    name: str
    type_: TestType
    status: TestStatus
    started_at: datetime | None
    finished_at: datetime | None
    duration_sec: float

    @classmethod
    def model_validate_test(cls, test: TestABC) -> Self:
        return cls.model_validate(
            {
                'id_': id(test),
                'name': test.name,
                'type_': test.type,
                'status': test.status,
                'started_at': test.started_at,
                'finished_at': test.finished_at,
                'duration_sec': test.duration_sec
            }
        )


class MultiTestInfo(TestInfoABC):
    """
    Multitest info model.
    """
    id_: int
    name: str
    type_: TestType
    status: TestStatus
    started_at: datetime | None
    finished_at: datetime | None
    duration_sec: float
    tests: list[TestInfo | MultiTestInfo]

    @classmethod
    def model_validate_test(cls, test: MultiTestABC) -> Self:
        return cls.model_validate(
            {
                'id_': id(test),
                'name': test.name,
                'type_': test.type,
                'status': test.status,
                'started_at': test.started_at,
                'finished_at': test.finished_at,
                'duration_sec': test.duration_sec,
                'tests': [
                    MultiTestInfo.model_validate_test(t)
                    if isinstance(t, MultiTestABC) else
                    TestInfo.model_validate_test(t)
                    for t in test
                ]
            }
        )



import pytest

from ._memory import TestMemory


@pytest.fixture
def test_memory():
    return TestMemory()


def test_new_key(test_memory):
    test_memory.new_key('key1', value=10)
    assert test_memory['key1'] == 10


def test_set_value(test_memory):
    test_memory.new_key('key1', value=10)
    test_memory['key1'] = 20
    assert test_memory['key1'] == 20


def test_del_item(test_memory):
    test_memory.new_key('key1', value=10)
    del test_memory['key1']
    with pytest.raises(KeyError):
        test_memory['key1']


def test_iter(test_memory):
    test_memory.new_key('key1', value=10)
    test_memory.new_key('key2', value=20)
    assert set(test_memory) == {'key1', 'key2'}


def test_len(test_memory):
    test_memory.new_key('key1', value=10)
    test_memory.new_key('key2', value=20)
    assert len(test_memory) == 2


@pytest.mark.asyncio
async def test_read(test_memory):
    test_memory.new_key('key1', value=10, history=True)
    test_memory['key1'] = 20
    test_memory['key1'] = 30
    assert await test_memory.read('key1') == 30
    assert await test_memory.read('key1', depth=2) == [10, 20, 30]


@pytest.mark.asyncio
async def test_write(test_memory):
    test_memory.new_key('key1', value=10)
    await test_memory.write('key1', 20)
    assert test_memory['key1'] == 20


@pytest.mark.asyncio
async def test_write_non_updatable(test_memory):
    test_memory.new_key('key1', value=10, updatable=False)
    with pytest.raises(KeyError):
        await test_memory.write('key1', 20)


@pytest.mark.asyncio
async def test_write_force_update(test_memory):
    test_memory.new_key('key1', value=10, updatable=False)
    await test_memory.write('key1', 20, force_update=True)
    assert test_memory['key1'] == 20



import asyncio

import shv

from g3shvnodes.route import SHVRouteControlNode
from g3shvnodes.gate import SHVGateControlNode
from g3shvnodes.detector import SHVDetectorControlNode, SHVDetectorStatusNode

from ...enums import TestStatus
from ...exceptions import TestFail, TestFailFatal
from ...tags import (
    command, before_command, value_getter, success, on_status_change
)
from ..._test import SingleTest


class RequestRouteTest(SingleTest[list[str]]):
    def __init__(
        self,
        route_node: SHVRouteControlNode,
        route_entry_gate_node: SHVGateControlNode,
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )
        self.route_node = route_node
        self.route_entry_gate_node = route_entry_gate_node
        self.gate_memory_before_request: list[str] = []

    @property
    def route_name(self) -> str:
        return self.route_node.path.split('/')[-1]

    @command
    async def request(self) -> shv.SHVType:
        return await self.route_node.request()

    @value_getter
    async def get_route_memory(self) -> list[str]:
        memory = await self.route_entry_gate_node.requestMemory.get()
        if not isinstance(memory, str):
            raise TestFailFatal(f'Expected str, got {type(memory).__name__}')
        return memory.splitlines()

    @before_command(command=request)
    async def get_route_memory_before_request(self) -> None:
        memory = await self.get_route_memory()
        self.gate_memory_before_request = memory

    @success
    def is_route_requested(self) -> bool:
        berofe_request = set(self.gate_memory_before_request)
        assert self.value is not None, 'value is None'
        after_request = set(self.value)
        diff = after_request - berofe_request
        route_name = self.route_name
        return any(route_name in line for line in diff)


class WaitUntilRouteIsReadyTest(SingleTest[int]):
    def __init__(
        self,
        route_control_node: SHVRouteControlNode,
        trigger_detector_control_node: SHVDetectorControlNode | None = None,
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        prevent_race_condition: bool = True
    ) -> None:
        super().__init__(
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=1
            )
        self.route_control_node = route_control_node
        self.trigger_detector_control_node = trigger_detector_control_node
        self.trigger_detector_occupied: bool = False
        self.trigger_detector_released: bool = False
        self.trigger_detector_status_update_task: asyncio.Task | None = None
        self.prevent_race_condition = prevent_race_condition

    @property
    def trigger_status_node(self) -> SHVDetectorStatusNode:
        if self.trigger_detector_control_node is None:
            raise AttributeError('trigger_detector_control_node is None')
        return self.trigger_detector_control_node.status

    @property
    def trigger_name(self) -> str:
        return self.trigger_status_node.path.split('/')[-2]

    @property
    def route_name(self) -> str:
        return self.route_control_node.path.split('/')[-1]

    async def _await_trigger_detector_occupied(self) -> int:
        self._logger.debug(
            'Waiting for the trigger detector "%s" to be occupied',
            self.trigger_name
            )
        status = await self.trigger_status_node.status_change_wait(
            any_false=[SHVDetectorStatusNode.BitAlias.FREE],
            any_true=[SHVDetectorStatusNode.BitAlias.OCCUPIED],
            timeout=self.max_duration,
            get_period=self._exec_sleep
            )
        self._logger.debug(
            'The trigger detector "%s" is occupied',
            self.trigger_name
            )
        return status

    async def _await_trigger_detector_released(self) -> int:
        self._logger.debug(
            'Waiting for the trigger detector "%s" to be released',
            self.trigger_name
            )
        status = await self.trigger_status_node.status_change_wait(
            any_false=[SHVDetectorStatusNode.BitAlias.OCCUPIED],
            any_true=[SHVDetectorStatusNode.BitAlias.FREE],
            timeout=self.max_duration,
            get_period=self._exec_sleep
            )
        self._logger.debug(
            'The trigger detector "%s" is released',
            self.trigger_name
            )
        return status

    async def track_trigger_detector(self) -> None:
        if self.trigger_detector_control_node is None:
            self.trigger_detector_occupied = True
            self.trigger_detector_released = True
            return
        await self._await_trigger_detector_occupied()
        self.trigger_detector_occupied = True
        await self._await_trigger_detector_released()
        self.trigger_detector_released = True

    @on_status_change(
        change_to=[TestStatus.RUNNING],
        priority=1
        )
    async def subscribe(self) -> None:
        if self.trigger_detector_control_node is not None:
            await self.trigger_status_node.subscribe()
        await self.route_control_node.status.subscribe()

    @on_status_change(
        change_to=[TestStatus.RUNNING],
        priority=2
        )
    def start_tracking_trigger_detector(self) -> None:
        if (
            self.trigger_detector_status_update_task is None or
            self.trigger_detector_status_update_task.done()
        ):
            self.trigger_detector_status_update_task = asyncio.create_task(
                self.track_trigger_detector()
                )

    @on_status_change(
        change_to=[
            TestStatus.PASSED,
            TestStatus.PASSED_WITH_WARNING,
            TestStatus.FAILED,
            TestStatus.TIMEOUT,
            TestStatus.ABORTED
            ],
        priority=1
        )
    async def stop_tracking_trigger_detector(self) -> None:
        if self.trigger_detector_status_update_task is not None:
            self.trigger_detector_status_update_task.cancel()
            try:
                await self.trigger_detector_status_update_task
            except asyncio.CancelledError:
                pass
            self.trigger_detector_status_update_task = None

    @on_status_change(
        change_to=[
            TestStatus.PASSED,
            TestStatus.PASSED_WITH_WARNING,
            TestStatus.FAILED,
            TestStatus.TIMEOUT,
            TestStatus.ABORTED
            ],
        priority=2
        )
    async def unsubscribe(self) -> None:
        if self.trigger_detector_control_node is not None:
            await self.trigger_status_node.unsubscribe()
        await self.route_control_node.status.unsubscribe()

    @value_getter
    async def get_route_status(self) -> int:
        return await self.route_control_node.status.get()

    @before_command
    async def await_route_status_change(self) -> None:
        route_status = await self.get_route_status()
        if self.route_control_node.status.is_free(route_status) is False:
            if self.trigger_detector_control_node is not None:
                raise TestFail(
                    f'Route "{self.route_name}" is not free in the beginning '
                    f'of the test. Route status: {route_status}. '
                    f'The trigger detector "{self.trigger_name}" state: '
                    f'FREED={self.trigger_detector_released}, '
                    f'OCCUPIED={self.trigger_detector_occupied}'
                    )
        self._logger.debug(
            'Waiting until the route "%s" switches to READY',
            self.route_name
            )
        bit_alias = self.route_control_node.status.BitAlias
        await self.route_control_node.status.status_change_wait(
            any_true=[bit_alias.READY, bit_alias.BUILD],
            any_false=[bit_alias.FREE],
            timeout=self.max_duration,
            get_period=self._exec_sleep
            )
        if self.trigger_detector_control_node is not None:
            # prevent race condition, when the trigger detector release after
            # its occupation is registered with a delay. This can happen if
            # the detector datachange event is not registered or
            # is registered with a delay. This is slightly unsafe, because
            # theoretically (although very unlikely) the detector release can
            # physically happen and be registered AFTER the route is ready,
            # which should fail the test, but will not.
            if self.prevent_race_condition:
                if (
                    self.trigger_detector_occupied is False or
                    self.trigger_detector_released is False
                ):
                    await asyncio.sleep(self._exec_sleep + 0.05)

    @success
    async def is_route_status_build_or_ready(self) -> bool:
        status = self.value
        assert status is not None, 'value has not been set by the getter'
        return (
            (
                self.route_control_node.status.is_build(status) or
                self.route_control_node.status.is_ready(status)
            ) and
            (
                self.trigger_detector_occupied is True and
                self.trigger_detector_released is True
            )
        )



import shv

from g3shvnodes.detector import SHVDetectorControlNode, SHVDetectorTestNode

from ..element import ControlNodeStatusValueSingleTest
from ...tags import command


class DetectorSetFreeTest(
    ControlNodeStatusValueSingleTest[SHVDetectorControlNode]
):
    def __init__(
        self,
        control_node: SHVDetectorControlNode,
        test_node: SHVDetectorTestNode,
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            node=control_node,
            success_value=control_node.status.BitAlias.FREE,
            fail_value=control_node.status.BitAlias.ERROR_INPUT,
            mask=control_node.status.BitMask.STATE,
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )
        self.test_node = test_node

    @command
    async def set_false(self) -> shv.SHVType:
        return await self.test_node.set_false()


class DetectorSetOccupiedTest(
    ControlNodeStatusValueSingleTest[SHVDetectorControlNode]
):
    def __init__(
        self,
        control_node: SHVDetectorControlNode,
        test_node: SHVDetectorTestNode,
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            node=control_node,
            success_value=control_node.status.BitAlias.OCCUPIED,
            fail_value=control_node.status.BitAlias.ERROR_INPUT,
            mask=control_node.status.BitMask.STATE,
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )
        self.test_node = test_node

    @command
    async def set_true(self) -> shv.SHVType:
        return await self.test_node.set_true()


class DetectorResetTest(
    ControlNodeStatusValueSingleTest[SHVDetectorControlNode]
):
    def __init__(
        self,
        control_node: SHVDetectorControlNode,
        test_node: SHVDetectorTestNode,
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            node=control_node,
            success_value=control_node.status.BitAlias.FREE,
            fail_value=control_node.status.BitAlias.ERROR_INPUT,
            mask=control_node.status.BitMask.STATE,
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )
        self.test_node = test_node

    @command
    async def reset(self) -> shv.SHVType:
        return await self.test_node.reset()



import enum


class TestStatus(enum.StrEnum):
    """Test status enumeration."""
    UNKNOWN = enum.auto()
    """Test status is unknown."""
    PENDING = enum.auto()
    """Test is pending for execution."""
    QUEUED = enum.auto()
    """Test is queued for execution."""
    RUNNING = enum.auto()
    """Test is running."""
    PASSED_WITH_WARNING = enum.auto()
    """Test passed with warning."""
    PASSED = enum.auto()
    """Test passed."""
    FAILED = enum.auto()
    """Test failed."""
    TIMEOUT = enum.auto()
    """Test timed out."""
    ABORTED = enum.auto()
    """Test was aborted."""


class TestType(enum.StrEnum):
    """Test type enumeration."""
    OTHER = enum.auto()
    """Other test type."""
    SINGLETEST = enum.auto()
    """Single test."""
    MULTITEST = enum.auto()
    """Multitest."""
    MULTITEST_SEQUENTIAL = enum.auto()
    """Sequential multitest."""
    MULTITEST_PARALLEL = enum.auto()
    """Parallel multitest."""



from ._route import RequestRouteTest, WaitUntilRouteIsReadyTest
from ._passage_generator import PassageGenerator

__all__ = [
    'RequestRouteTest',
    'WaitUntilRouteIsReadyTest',
    'PassageGenerator'
]



import typing

import shv

from g3shvnodes.node import SHVNode
from g3shvnodes.pointmachine import (
    SHVPMEControlNode,
    SHVPMMControlNode,
    SHVPMETestNode,
    SHVPMMTestNode
)

from ...exceptions import TestFailFatal
from ...tags import command, after_command
from ..element import ControlNodeStatusValueSingleTest


class _PMPositionTest(
    ControlNodeStatusValueSingleTest[SHVPMEControlNode | SHVPMMControlNode]
):
    def __init__(
        self,
        success_value: int,
        control_node: SHVPMEControlNode | SHVPMMControlNode,
        test_node: SHVPMETestNode | SHVPMMTestNode | None = None,
        pm_type: typing.Literal['pme', 'pmm', 'auto'] = 'auto',
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            node=control_node,
            success_value=success_value,
            fail_value=control_node.status.BitAlias.ERROR_POSITION,
            mask=control_node.status.BitMask.POSITION,
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )
        self.test_node = test_node
        if pm_type.lower() == 'auto':
            self.pm_type = self.get_pm_type_from_node_type(control_node)
        else:
            self.pm_type = pm_type.lower()

    @staticmethod
    def get_pm_type_from_node_type(node: SHVNode) -> str:
        if isinstance(node, SHVPMEControlNode):
            return 'pme'
        if isinstance(node, SHVPMMControlNode):
            return 'pmm'
        raise ValueError(f'Unknown node type: "{type(node).__name__}"')

    @after_command
    async def await_motor_not_moving(self) -> bool:
        if self.pm_type == 'pmm':
            return True
        if not hasattr(self.control_node.status.BitAlias, 'MOTOR_MOVING'):
            self._logger.warning(
                '"MOTOR_MOVING" bit was not found in status BitAlias enum. '
                'Skipping check.'
                )
            return True
        await self.control_node.status.status_change_wait(
            any_false=[self.control_node.status.BitAlias.MOTOR_MOVING],
            timeout=self.max_duration,
            get_period=self._exec_sleep
            )
        return True


class PMSetPositionTest(_PMPositionTest):
    def __init__(
        self,
        control_node: SHVPMEControlNode | SHVPMMControlNode,
        test_node: SHVPMETestNode | SHVPMMTestNode | None = None,
        set_to: typing.Literal['left', 'right', 'middle'] = 'left',
        set_from: typing.Literal['test', 'control'] = 'test',
        pm_type: typing.Literal['pme', 'pmm', 'auto'] = 'auto',
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            success_value=self.pos_to_bit(control_node, set_to),
            control_node=control_node,
            test_node=test_node,
            pm_type=pm_type,
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )
        self.set_to = set_to.lower()
        self.set_from = set_from.lower()

    @staticmethod
    def pos_to_bit(
        node: SHVPMEControlNode | SHVPMMControlNode, set_to_pos: str
    ) -> int:
        set_to_pos = set_to_pos.lower()
        if set_to_pos == 'left':
            return node.status.BitAlias.LEFT
        if set_to_pos == 'right':
            return node.status.BitAlias.RIGHT
        if set_to_pos == 'middle':
            return node.status.BitAlias.MIDDLE
        raise ValueError(f'Invalid position: "{set_to_pos}"')

    @command
    async def set_position(self) -> shv.SHVType:
        match self.set_from, self.pm_type, self.set_to:
            case 'test', _, 'left':
                assert self.test_node is not None, 'Test node not set'
                return await self.test_node.set_position_left()
            case 'test', _, 'right':
                assert self.test_node is not None, 'Test node not set'
                return await self.test_node.set_position_right()
            case 'test', _, 'middle':
                assert self.test_node is not None, 'Test node not set'
                return await self.test_node.set_position_middle()
            case 'control', 'pme', 'left':
                if not hasattr(self.control_node, 'switch_left'):
                    node_cls_name = type(self.control_node).__name__
                    raise TestFailFatal(
                        f'PME control node "{node_cls_name}" doe not '
                        f'implement "switch_left" method'
                        )
                return await self.control_node.switch_left()
            case 'control', 'pme', 'right':
                if not hasattr(self.control_node, 'switch_right'):
                    node_cls_name = type(self.control_node).__name__
                    raise TestFailFatal(
                        f'PME control node "{node_cls_name}" doe not '
                        f'implement "switch_right" method'
                        )
                return await self.control_node.switch_right()
            case _:
                raise TestFailFatal(
                    f'Invalid combination of test parameters: '
                    f'set_from="{self.set_from}", pm_type="{self.pm_type}", '
                    f'set_to="{self.set_to}"'
                    )


class PMResetTest(_PMPositionTest):
    def __init__(
        self,
        control_node: SHVPMEControlNode | SHVPMMControlNode,
        test_node: SHVPMETestNode | SHVPMMTestNode,
        pm_type: typing.Literal['pme', 'pmm', 'auto'] = 'auto',
        name: str | None = None,
        execution_sleep: float = 1,
        min_duration: float | None = None,
        max_duration: float | None = None,
        max_evaluation_retries: int | None = None
    ) -> None:
        super().__init__(
            success_value=control_node.status.BitAlias.LEFT,
            control_node=control_node,
            test_node=test_node,
            pm_type=pm_type,
            name=name,
            execution_sleep=execution_sleep,
            min_duration=min_duration,
            max_duration=max_duration,
            max_evaluation_retries=max_evaluation_retries
            )

    @command
    async def reset(self) -> shv.SHVType:
        assert self.test_node is not None, 'Test node not set'
        return await self.test_node.reset()




import math
import typing
import logging

from shv import ValueClient

from g3elements import (
    ElementABC,
    Detector,
    PointMachineElectrical,
    PointMachineMechanical,
    Route,
    Gate,
    Zone,
    Signal,
    SignalSymbol,
    PointMachine,
    RouteLayoutDetector,
    ConnectorNotSetError,
    PointMachinePosition,
)
from g3elements.route_utils import RouteAnalyzer, DetectorEvent, RouteManager

from g3shvnodes import SHVNode
from g3shvnodes.detector import SHVDetectorControlNode, SHVDetectorTestNode
from g3shvnodes.pointmachine import (
    SHVPMEControlNode, SHVPMETestNode, SHVPMMControlNode, SHVPMMTestNode
)
from g3shvnodes.route import SHVRouteControlNode, SHVRouteStatusNode
from g3shvnodes.gate import SHVGateControlNode
from g3shvnodes.zone import SHVZoneControlNode
from g3shvnodes.signal import (
    SHVSignalControlNode,
    SHVSignalSymbolControlNode,
    SHVSignalSymbolTestNode
)

from ..._test import (
    TestABC,
    MultiTestABC,
    ParallelMultiTest,
    SequentialMultiTest
)
from ..detector import DetectorSetOccupiedTest, DetectorSetFreeTest
from ..element import ControlNodeStatusValueSingleTest
from ..zone import ResetZoneTest
from ..route import RequestRouteTest, WaitUntilRouteIsReadyTest


logger = logging.getLogger('g3tester')


TControlNode = typing.TypeVar('TControlNode', bound=typing.Optional[SHVNode])
TTestNode = typing.TypeVar('TTestNode', bound=typing.Optional[SHVNode])


class ElementSHVNodes(
    typing.NamedTuple, typing.Generic[TControlNode, TTestNode]
):
    control: TControlNode
    test: TTestNode


class SHVNodeFactory:
    def __init__(
        self,
        control_client: ValueClient,
        test_client: ValueClient,
        control_paths: dict[str, str] | None = None,
        test_paths: dict[str, str] | None = None
    ) -> None:
        self.control_client = control_client
        self.test_client = test_client
        self.control_paths = control_paths or {}
        self.test_paths = test_paths or {}

    def get_default_path_control(self, element: ElementABC) -> str:
        return self.control_paths.setdefault(
            element.shv_path_full, f'shv/{element.shv_path_full}'
            )

    def get_default_path_test(self, element: ElementABC) -> str:
        return self.test_paths.setdefault(
            element.shv_path_full, f'shv/{element.shv_path_full}/_test'
            )

    @typing.overload
    def __call__(
        self, element: Detector
    ) -> ElementSHVNodes[SHVDetectorControlNode, SHVDetectorTestNode]: ...

    @typing.overload
    def __call__(
        self, element: PointMachineElectrical
    ) -> ElementSHVNodes[SHVPMEControlNode, SHVPMETestNode]: ...

    @typing.overload
    def __call__(
        self, element: PointMachineMechanical
    ) -> ElementSHVNodes[SHVPMMControlNode, SHVPMMTestNode]: ...

    @typing.overload
    def __call__(
        self, element: Gate
    ) -> ElementSHVNodes[SHVGateControlNode, None]: ...

    @typing.overload
    def __call__(
        self, element: Route
    ) -> ElementSHVNodes[SHVRouteControlNode, None]: ...

    @typing.overload
    def __call__(
        self, element: Signal
    ) -> ElementSHVNodes[SHVSignalControlNode, None]: ...

    @typing.overload
    def __call__(
        self, element: SignalSymbol
    ) -> ElementSHVNodes[SHVSignalControlNode, SHVSignalSymbolTestNode]: ...

    @typing.overload
    def __call__(
        self, element: Zone
    ) -> ElementSHVNodes[SHVZoneControlNode, None]: ...

    def __call__(self, element: ElementABC) -> ElementSHVNodes:
        control_node_cls: typing.Type[SHVNode] | None
        test_node_cls: typing.Type[SHVNode] | None
        if isinstance(element, Detector):
            control_node_cls = SHVDetectorControlNode
            test_node_cls = SHVDetectorTestNode
        elif isinstance(element, PointMachineElectrical):
            control_node_cls = SHVPMEControlNode
            test_node_cls = SHVPMETestNode
        elif isinstance(element, PointMachineMechanical):
            control_node_cls = SHVPMMControlNode
            test_node_cls = SHVPMMTestNode
        elif isinstance(element, Gate):
            control_node_cls = SHVGateControlNode
            test_node_cls = None
        elif isinstance(element, Route):
            control_node_cls = SHVRouteControlNode
            test_node_cls = None
        elif isinstance(element, Signal):
            control_node_cls = SHVSignalControlNode
            test_node_cls = None
        elif isinstance(element, SignalSymbol):
            control_node_cls = SHVSignalSymbolControlNode
            test_node_cls = SHVSignalSymbolTestNode
        elif isinstance(element, Zone):
            control_node_cls = SHVZoneControlNode
            test_node_cls = None
        else:
            raise NotImplementedError(
                f'Element type "{type(element).__name__}" is not supported.'
                )
        if control_node_cls is not None:
            control_path = self.get_default_path_control(element)
            control_node = control_node_cls(control_path)
            control_node.set_client_unsafe(self.control_client)
        else:
            control_node = None
        if test_node_cls is not None:
            test_path = self.get_default_path_test(element)
            test_node = test_node_cls(test_path)
            test_node.set_client_unsafe(self.test_client)
        else:
            test_node = None
        return ElementSHVNodes(control_node, test_node)


class PassageGenerator:
    _MIN_TEST_MAX_DURATION: float = 2.0
    _PM_SWITCH_DURATION: float = 3.0
    _DETECTOR_RELEASE_DURATION: float = 2.0
    _MAX_DURATION_MULTIPLIER: float = 1.5

    def __init__(
        self,
        *routes: Route | typing.Iterable[Route | typing.Iterable],
        control_client: ValueClient,
        test_client: ValueClient,
        control_paths: dict[str, str] | None = None,
        test_paths: dict[str, str] | None = None,
    ) -> None:
        self.route_analyzer = RouteAnalyzer(*routes)
        self.shvnode_factory = SHVNodeFactory(
            control_client, test_client, control_paths, test_paths
            )

    @property
    def route_manager(self) -> RouteManager:
        return self.route_analyzer._route_manager

    @staticmethod
    def _check_if_above_boundary(
        value: int | float, boundary: int | float = 0
    ):
        """
        Check if the given value is strictly greater than a specified boundary.

        Args:
            value (int | float): The value to check.
            boundary (int | float): The lower limit that `value` must exceed.

        Raises:
            TypeError: If the value is not a number.
            ValueError: If the value is not strictly greater than the boundary.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(f'"{value}" is not a numeric value.')
        if value <= boundary:
            raise ValueError(f'Value must be greater than {boundary}.')

    @classmethod
    def set_minimum_test_max_duration(cls, max_duration: float = 2.0):
        """
        Sets a minimum base `max_duration` for a test. Any test with a finite
        `max_duration` will have the `max_duration` value at least as large as
        this minimum `max_duration` value, adjusted by the maximum duration
        multiplier factor (set with the `set_max_duration_multiplier` method).
        If a test has an infinite duration, this base value does not affect it.

        Args:
            max_duration (float): The minimum base `max_duration` value
            in seconds. Default is 2.0 seconds.

        Raises:
            ValueError: If max_duration is not strictly greater than 0.
            TypeError: If the max_duration is not a numeric value.
        """
        cls._check_if_above_boundary(max_duration)
        cls._MIN_TEST_MAX_DURATION = max_duration

    @classmethod
    def set_pm_switch_duration(cls, duration: float = 3.0):
        """
        Sets the duration of a point machine switch from one position
        to another.

        Args:
            duration (float): The PM switch duration in seconds.
            Default is 3.0 seconds.

        Raises:
            ValueError: If duration is not strictly greater than 0.
            TypeError: If the duration is not a numeric value.
        """
        cls._check_if_above_boundary(duration)
        cls._PM_SWITCH_DURATION = duration

    @classmethod
    def set_detector_release_duration(cls, duration: float = 2.0):
        """
        Sets the duration of a position detector release lag.

        Args:
            duration (float): The detector release duration in seconds.
            Default is 2.0 seconds.

        Raises:
            ValueError: If duration is not strictly greater than 0.
            TypeError: If the duration is not a numeric value.
        """
        cls._check_if_above_boundary(duration)
        cls._DETECTOR_RELEASE_DURATION = duration

    @classmethod
    def set_max_duration_multiplier(cls, multiplier: float = 1.5):
        """
        Sets the safety multiplier for a test max_duration.

        Args:
            multiplier (float): The max_duration multiplier.
            Must be greater than 1. Default is 2.0.

        Raises:
            ValueError: If multiplier is not strictly greater than 1.
            TypeError: If the multiplier is not a numeric value.
        """
        cls._check_if_above_boundary(multiplier, boundary=1)
        cls._MAX_DURATION_MULTIPLIER = multiplier

    def _ceil_with_multiplier(self, base_max_duration: float) -> float:
        """
        Multiplies the given duration value with maximum duration
        multiplier and rounds it up to the next natural number.

        Args:
            base_max_duration (float): The base max_duration value to be
            multiplied.

        Returns:
            float: The multiplied max_duration value rounded up.

        Raises:
            ValueError: If base_max_duration is not strictly greater than 0.
            TypeError: If base_max_duration is not a numeric value.
        """
        self._check_if_above_boundary(base_max_duration)
        return math.ceil(self._MAX_DURATION_MULTIPLIER * base_max_duration)

    def _calculate_pms_switch_duration(
        self, pms: typing.Sequence[PointMachine]
    ) -> float:
        """
        Calculates the total time required for all point machines to
        switch. The calculation is based on the assumption that only one
        point machine can switch at a time, which is a worst-case scenario.

        Args:
            pms (list[PointMachineType | str]): List of point machines.

        Returns:
            float: The time required for all point machines to switch.
        """
        return self._PM_SWITCH_DURATION * len(pms)

    def _calculate_detectors_release_duration(
        self,
        detectors: typing.Sequence[Detector],
        last_detector: Detector | None = None
    ) -> float:
        """
        Calculates the total time required to release a sequence of detectors,
        under the assumption that only one detector can be released at a time
        (a worst-case scenario). This method allows for truncating
        the sequence at a specified `last_detector`, which can be useful in
        truncated route passage scenarios.

        Args:
            detectors (list[DetectorType | str]): List of detectors.
            last_detector (DetectorType | None): The last detector to be
            released in the sequence. If None, the duration considers
            the release of all detectors in the list.

        Returns:
            float: The total time required to release the specified detectors.
        """
        duration: float = 0
        for det in detectors:
            duration += self._DETECTOR_RELEASE_DURATION
            if det == last_detector:
                break
        return duration

    def _calculate_zone_reset_duration(self, zone: Zone) -> float:
        """
        Calculates the total time required to reset a zone.
        It is determined by the maximum of either the time to switch all
        point machines or the time to release all detectors in the zone.

        Args:
            zone (Zone): The zone to be reset.

        Returns:
            float: The total time required to reset the zone.
        """
        pmes = zone.find_by_type(PointMachineElectrical)
        pmms = zone.find_by_type(PointMachineMechanical)
        dets = zone.find_by_type(Detector)
        return max(
            self._calculate_pms_switch_duration(pmes + pmms),  # type: ignore
            self._calculate_detectors_release_duration(dets)
            )

    def _calculate_route_set_duration(self, route: Route) -> float:
        """
        Calculates the total time required to set a route as the time
        it takes to switch all the point machines in the route layout.

        Args:
            route (Route): The route to be set.

        Returns:
            float: The total time required to set a Route.
        """
        pms = [pm.core for pm in route.layout.pointmachines]
        return self._calculate_pms_switch_duration(pms)

    def _calculate_total_test_max_duration(self, test: TestABC) -> float:
        """
        Recursively calculates the total maximum duration for a given test.
        This method takes into account the hierarchical structure of tests,
        which can include MultiTests that contain child tests.

        For a test with a finite `max_duration`, it will simply return
        this `max_duration`, regardless of whether the test is a `MultiTest`
        or not. This is to avoid  potentially double-counting the child tests'
        maximum durations.

        For a `MultiTest` with an infinite `max_duration`, the method behaves
        differently depending on the type of this `MultiTest`:
        - `SequentialMultiTest`: returns the sum of the maximum durations of
        all child tests.
        - `ParallelMultiTest`: returns the biggest maximum duration among
        all child tests.

        Args:
            test (TestABC): The test for which the total maximum duration
            is calculated.

        Returns:
            float: The calculated total max_duration for the test.
        """
        if test.max_duration != float('inf'):
            return test.max_duration
        if isinstance(test, MultiTestABC):
            child_max_durations = [
                self._calculate_total_test_max_duration(t) for t in test
                ]
            if any(d == float('inf') for d in child_max_durations):
                return float('inf')
            if isinstance(test, SequentialMultiTest):
                return sum(child_max_durations)
            if isinstance(test, ParallelMultiTest):
                return max(child_max_durations, default=float('inf'))
        return float('inf')

    def _calculate_time_between_events(
        self, event: DetectorEvent, i_event: int, events: list[DetectorEvent]
    ) -> float:
        """
        Calculates the time between the given event and the next event in the
        list. This time is the time it takes the passing tram to move between
        the position detector areas. If the given event is the last one in
        the list, it calculates the time until the tram leaves the exit gate.

        Args:
            event (DetectorEvent): The initial event.
            i_event (int): The index of the initial event in the events list.
            events (list[DetectorEvent]): The list of events.

        Returns:
            float: The time between the initial event and the next event or
            exit time.
        """
        try:
            next_event = events[i_event + 1]
            return next_event.time - event.time
        except IndexError:
            next_event_time = self.route_analyzer.estimate_passage_duration(
                route=event.detector.layout.route
                )  # tram leaves the exit gate
            return next_event_time - event.time

    def _assert_zone_set(self, zone: Zone | None) -> typing.TypeGuard[Zone]:
        """
        Asserts that the provided zone object is not None.

        Args:
            zone (Zone | None): The zone object to be checked.

        Raises:
            ConnectorNotSetError: If the zone object is None.
        """
        if zone is None:
            raise ConnectorNotSetError('Zone element has not been found.')
        return True

    def _generate_reset_zone_test(self, zone: Zone) -> ResetZoneTest:
        """
        Generates a `ResetZone` test instance. This test applier the "reset"
        command to every point machine and position detector in the given zone.

        Args:
            zone (Zone | typing.Any): The zone to be reset.

        Raises:
            InvalidZone: If the provided zone object is None or
            not an instance of the `Zone` object.

        Returns:
            ResetZone: A `ResetZone` test instance.
        """
        duration = self._calculate_zone_reset_duration(zone)
        detectors = zone.find_by_type(Detector)
        pmes = zone.find_by_type(PointMachineElectrical)
        pmms = zone.find_by_type(PointMachineMechanical)
        zone_nodes = self.shvnode_factory(zone)
        detectors_nodes = [self.shvnode_factory(det) for det in detectors]
        pmes_nodes = [self.shvnode_factory(pm) for pm in pmes]
        pmms_nodes = [self.shvnode_factory(pm) for pm in pmms]
        return ResetZoneTest(
            zone_control_node=zone_nodes.control,
            detector_control_nodes=[det.control for det in detectors_nodes],
            detector_test_nodes=[det.test for det in detectors_nodes],
            pm_control_nodes=[pm.control for pm in pmes_nodes + pmms_nodes],
            pm_test_nodes=[pm.test for pm in pmes_nodes + pmms_nodes],
            name=f'Reset zone {zone.name}',
            min_duration=None,
            max_duration=self._ceil_with_multiplier(duration)
            )

    def _generate_request_route_test(self, route: Route) -> RequestRouteTest:
        """
        Generates a `RequestRoute` test instance.
        This test is responsible for requesting a route for a passage.

        Args:
            route (Route): The route to be requested.

        Returns:
            RequestRoute: A `RequestRoute` test instance.
        """
        duration = self._MIN_TEST_MAX_DURATION
        gate = route.layout.entry_gate
        return RequestRouteTest(
            route_node=self.shvnode_factory(route).control,
            route_entry_gate_node=self.shvnode_factory(gate).control,
            name=f'Request route {route.name}',
            min_duration=None,
            max_duration=self._ceil_with_multiplier(duration),
            )

    def _generate_check_detector_claimed_test(
        self, detector: Detector
    ) -> ControlNodeStatusValueSingleTest:
        node = self.shvnode_factory(detector).control
        return ControlNodeStatusValueSingleTest(
            node=node,
            success_value=node.status.BitAlias.CLAIMED,
            fail_value=node.status.BitAlias.ERROR_INPUT,
            mask=node.status.BitMask.CLAIMED,
            name=f'Check if detector {detector.name} is CLAIMED',
            max_duration=4,
            max_evaluation_retries=1
            )

    def _generate_check_pm_position_test(
        self, pm: PointMachine, position: PointMachinePosition
    ) -> ControlNodeStatusValueSingleTest:
        node = self.shvnode_factory(pm).control
        match position:
            case PointMachinePosition.LEFT:
                position_value = node.status.BitAlias.LEFT
            case PointMachinePosition.RIGHT:
                position_value = node.status.BitAlias.RIGHT
            case PointMachinePosition.MIDDLE:
                position_value = node.status.BitAlias.MIDDLE
            case _:
                raise ValueError(f'Invalid position value: {position}')
        return ControlNodeStatusValueSingleTest(
            node=node,
            success_value=position_value,
            fail_value=node.status.BitAlias.ERROR_POSITION,
            mask=node.status.BitMask.POSITION,
            name=f'Check if PM {pm.name} is in {position.name}',
            max_duration=4,
            max_evaluation_retries=1
            )

    def _generate_wait_for_route_test(
        self,
        route: Route,
        trigger: RouteLayoutDetector | None,
        check_elements_before_passage: bool,
        min_duration: float | None = None,
        max_duration_offset: float = 0.0
    ) -> WaitUntilRouteIsReadyTest | SequentialMultiTest:
        """
        Generates a `WaitForRoute` test. If `duration` is a non-zero value,
        the test is wrapped in a `SequentialMultiTest` with a wait period.
        This test is responsible for waiting for a route's trigger detector
        to be released.

        Args:
            route (Route): The route to be awaited for.
            trigger (RouteDetector | None): The detector that must be freed
            before this route may be passed.
            check_elements_before_passage (bool): Flag indicating if the state
            of the route's point machines and detectors should be checked
            before the passage.
            duration (float, optional): Optional wait period. Default is 0.0.
            max_duration_offset (float, optional): An offset to add to the minimum  # noqa
            max_duration of the `WaitForRoute` test to reflect the waiting time for 
            the previous routes to pass.

        Returns:
            Union[WaitForRoute, SequentialMultiTest]: The test instance.
        """
        max_duration = self._calculate_route_set_duration(route)
        if trigger is None:
            trigger_detector_node = None
            name = f'Wait until route {route.name} is READY'
        else:
            trigger_detector_node = self.shvnode_factory(trigger.core).control
            name = (
                f'Wait until route {route.name} is READY '
                f'after detector {trigger.core.name} is released'
                )
        wait_for_route_test = WaitUntilRouteIsReadyTest(
            route_control_node=self.shvnode_factory(route).control,
            trigger_detector_control_node=trigger_detector_node,
            name=name,
            min_duration=None,
            max_duration=(max_duration + max_duration_offset),
            prevent_race_condition=True
            )
        if not check_elements_before_passage:
            return wait_for_route_test
        det_tests = [
            self._generate_check_detector_claimed_test(det.core)
            for det in route.layout.detectors
            ]
        pm_tests = [
            self._generate_check_pm_position_test(pm.core, pm.position)
            for pm in route.layout.pointmachines
            ]
        return SequentialMultiTest(
            wait_for_route_test,
            ParallelMultiTest(
                *det_tests,
                *pm_tests,
                name=f'Check elements before route {route.name} passage'
                ),
            name=f'Wait for route {route.name} and check elements'
            )

    def _generate_route_state_check_test(
        self, route: Route, released_detector: RouteLayoutDetector
    ) -> ControlNodeStatusValueSingleTest:
        """
        Generates a 'CheckIfRouteFree' or 'CheckIfRouteOccupied' test instance
        based on the position of the released detector in the route. The test
        is used to check if the route is free or occupied after a detector
        has been released.

        Args:
            route (Route): The route to be checked.
            released_detector (RouteDetector): The detector that has been
            released.

        Returns:
            Union[CheckIfRouteFree, CheckIfRouteOccupied]: The test instance.
        """
        route_node = self.shvnode_factory(route).control
        if route.layout.index(released_detector) == len(route.layout) - 1:
            name = (
                f'Check if route {route.name} is FREE after detector '
                f'{released_detector.name} is released'
                )
            return ControlNodeStatusValueSingleTest(
                node=route_node,
                success_value=SHVRouteStatusNode.BitAlias.FREE,
                fail_value=SHVRouteStatusNode.BitAlias.OCCUPIED,
                mask=SHVRouteStatusNode.BitMask.STATE,
                name=name,
                execution_sleep=1,
                max_duration=4,
                max_evaluation_retries=1
                )
        else:
            name = (
                f'Check if route {route.name} is OCCUPIED after detector '
                f'{released_detector.name} is released'
                )
            return ControlNodeStatusValueSingleTest(
                node=route_node,
                success_value=SHVRouteStatusNode.BitAlias.OCCUPIED,
                fail_value=SHVRouteStatusNode.BitAlias.FREE,
                mask=SHVRouteStatusNode.BitMask.STATE,
                name=f'Check route {route.name} status',
                max_duration=4,
                max_evaluation_retries=1
                )

    def _generate_detector_occupation_test(
        self,
        detector: RouteLayoutDetector,
        min_duration: float | None = None
    ) -> DetectorSetOccupiedTest:
        """
        Generates a `SetDetectorOccupied` test, which simulates the occupation
        of a position detector. If `duration` is a non-zero value, the test is
        wrapped in a `ParallelMultiTest` with a wait period, which represents
        the time it takes a passing tram to move between the detector areas.

        Args:
            detector (RouteDetector): The detector to be set as occupied.
            duration (float, optional): Time for the Wait test to simulate
            the tram's movement time. Defaults to 0.

        Returns:
            Union[SetDetectorOccupied, ParallelMultiTest]: The generated test.
        """
        max_duration = self._ceil_with_multiplier(self._MIN_TEST_MAX_DURATION)
        return DetectorSetOccupiedTest(
            control_node=self.shvnode_factory(detector.core).control,
            test_node=self.shvnode_factory(detector.core).test,
            name=f'Set {detector.core.name} occupied',
            min_duration=None,
            max_duration=max_duration
            )

    def _generate_detector_release_test(
        self,
        detector: RouteLayoutDetector,
        route: Route,
        duration: float | None = None
    ) -> SequentialMultiTest:
        """
        Generates a `SetDetectorFree` test, which simulates the occupation
        of a position detector. If `duration` is a non-zero value, the test is
        wrapped in a `ParallelMultiTest` with a wait period, which represents
        the time it takes a passing tram to move between the detector areas.

        Args:
            detector (RouteDetector): The detector to be set as occupied.
            duration (float, optional): Time for the Wait test to simulate
            the tram's movement time. Defaults to 0.

        Returns:
            Union[SetDetectorOccupied, ParallelMultiTest]: The generated test.
        """
        # create a test to change the detector state
        max_duration = self._ceil_with_multiplier(self._MIN_TEST_MAX_DURATION)
        set_detector_free = DetectorSetFreeTest(
            control_node=self.shvnode_factory(detector.core).control,
            test_node=self.shvnode_factory(detector.core).test,
            name=f'Set {detector.core.name} free',
            min_duration=None,
            max_duration=max_duration
            )
        # create a test to validate the route state after the change
        check_route = self._generate_route_state_check_test(route, detector)
        # combine the tests into a SequentialMultiTest
        return SequentialMultiTest(
            set_detector_free,
            check_route,
            name=f'Release {detector.name} and check route {route.name} state'
            )

    def _generate_detector_state_change_test(
        self,
        detector: RouteLayoutDetector,
        change_state_to: bool,
        route: Route,
        duration: float | None = None
    ) -> typing.Union[DetectorSetOccupiedTest, SequentialMultiTest]:
        """
        Generates a test or test sequence to simulate a state change of
        a detector, either to 'occupied' or 'free'. Depending on the
        `change_state_to` parameter, this function calls either
        `_generate_detector_occupation_test` or
        `_generate_detector_release_test`.

        Args:
            detector (RouteDetector): The detector, the state of which is
            to be changed.
            change_state_to (bool): The desired state of the detector.
            True for 'occupied', False for 'free'.
            route (Route): The route for the route state check test in case of
            detector release.
            duration (float, optional): Time for the Wait test to simulate
            the tram's  movement time. Defaults to 0.

        Returns:
            Union[SetDetectorOccupied, SequentialMultiTest, ParallelMultiTest]:
            The generated test or test sequence.
        """
        if change_state_to is True:  # detector occupation event
            return self._generate_detector_occupation_test(
                detector, duration
                )
        elif change_state_to is False:  # detector release event
            return self._generate_detector_release_test(
                detector, route, duration
                )
        raise TypeError(
            f'Invalid value for the "change_state_to" parameter: '
            f'{change_state_to}'
            )

    def _generate_passage_test_sequence(
        self,
        route: Route,
        fastmode: bool,
        trigger: RouteLayoutDetector | None,
        last_detector_to_release: RouteLayoutDetector | None,
        check_elements_before_passage: bool,
        wait_for_route_max_duration_offset: float = 0.0
    ) -> SequentialMultiTest:
        """
        This function creates a sequence of tests that simulates the passage of
        a tram through the specified route. It constructs this simulation based
        on the trigger detector of the route and the sequence of the occupation
        and release events of the detectors if the route layout
        (up to the specified last detector).

        Args:
            route (Route): The route to be passed.
            fastmode (bool): If True, the time delays between the detector
            occupation and release events in the passage test sequence are
            ignored.
            trigger (RouteDetector | None): A detector that needs to be
            released before the route switches to BUILD / READY state.
            If None, it means that the route can start immediately, which is
            common for the first route in the passage sequence or for a route
            that does not share elements with any passing route.
            last_detector_to_release (RouteDetector): This is the last
            detector in the layout of the route that needs to be released
            as part of the test sequence.
            check_elements_before_passage (bool): A flag indicating whether
            the state of point machines and detectors in the layout of the
            route should be checked before the passage test sequence begins.
            wait_for_route_max_duration_offset (float, optional): An additional
            time delay added to the max_duration for the `WaitForRoute` test.
            It is used when there is a need to account for the passage time of
            previous routes. Defaults to 0.0.

        Returns:
            SequentialMultiTest: The passage test sequence.
        """
        passage_test_max_duration: float = 0
        passage_test_sequence: list[TestABC] = []
        events = self.route_analyzer._determine_detector_events(
            route, last_detector_to_release
            )
        # the time it takes the front of the tram to get to
        # the start of the first detector area is the route layout
        time_to_first_detector: float = 0 if fastmode else events[0].time
        # start the passage test sequence with checking the route state
        wait_for_route_test = self._generate_wait_for_route_test(
            route, trigger, check_elements_before_passage,
            time_to_first_detector, wait_for_route_max_duration_offset
            )
        passage_test_sequence.append(wait_for_route_test)
        passage_test_max_duration += time_to_first_detector
        passage_test_max_duration += self._calculate_total_test_max_duration(
            wait_for_route_test
            )
        # iterate though the events and generate the corresponding tests
        for i, event in enumerate(events):
            if fastmode:
                time_between_events: float = 0
            else:
                time_between_events = self._calculate_time_between_events(
                    event, i, events
                    )
            # create a detector test based on the event type
            detector_test = self._generate_detector_state_change_test(
                event.detector, event.action, route, time_between_events
                )
            detector_test_duration = self._calculate_total_test_max_duration(
                detector_test
                )
            passage_test_max_duration += time_between_events
            passage_test_max_duration += detector_test_duration
            passage_test_sequence.append(detector_test)
        return SequentialMultiTest(
            *passage_test_sequence,
            max_duration=math.ceil(passage_test_max_duration),
            name=f'Route {route.name} passage simulation'
            )  # wrap the generated tests into a SequentialMultiTest

    def _generate_passage_scenario_single_route(
        self,
        route: Route,
        add_zone_reset: bool,
        truncate: bool,
        fastmode: bool,
    ) -> SequentialMultiTest:
        """
        Generate a tram passage scenario for a single route
        with a route request and an optional Zone reset.

        The scenario includes:

        1. `ResetZone` (optional);
        2. `RequestRoute`;
        3. `WaitForRoute` (checks the state of the route and its elements);
        4. A sequence of `SetDetectorOccupied` and `SetDetectorFree` tests
        (the latter are applied in conjuction with the route state ckecker
        tests `CheckIfRouteOccupied` or `CheckIfRouteFree`).

        Args:
            route (Route): The Route object to generate the scenario for.
            add_zone_reset (bool): If True, add a Zone reset test sequence,
            which resets all the detectors the point machines in this Zone.
            truncate (bool): If True, the passage scenario ends with the
            occupation of the route's first detector. Otherwise the scenario
            continues to the release of the route's last detector.
            fastmode (bool): If True, the time delays between the detector
            occupation and release events in the passage test sequence are
            ignored.

        Returns:
            SequentialMultiTest: The passage scenario wrapped into
            a `SequentialMultiTest`.
        """
        max_duration: float = 0
        tests: list[TestABC] = []
        # add the Reset test (optional)
        if add_zone_reset:
            if route.zone is None:
                raise ValueError(f'Route {route.name} has no Zone assigned.')
            reset_zone_test = self._generate_reset_zone_test(route.zone)
            tests.append(reset_zone_test)
            max_duration += self._calculate_total_test_max_duration(
                reset_zone_test
                )
        # add the RouteRequest test
        request_route_test = self._generate_request_route_test(route)
        tests.append(request_route_test)
        max_duration += self._calculate_total_test_max_duration(
            request_route_test
            )
        # create a passage test sequence
        last_detector_to_release = None if truncate else route.layout.detectors[-1]  # noqa
        passage_test = self._generate_passage_test_sequence(
            route=route,
            fastmode=fastmode,
            trigger=None,
            last_detector_to_release=last_detector_to_release,
            check_elements_before_passage=True,
            )
        tests.append(passage_test)
        max_duration += self._calculate_total_test_max_duration(passage_test)
        return SequentialMultiTest(
            *tests,
            max_duration=max_duration,
            name=f'Route {route.name} passage'
            )

    def _generate_passage_scenario_multiple_routes(
        self,
        routes: list[Route],
        add_zone_reset: bool,
        truncate: bool,
        fastmode: bool,
    ) -> SequentialMultiTest:
        """
        Generate a tram passage scenario for multiple routes with
        a route request for each route and an optional Zone reset.

        The scenario includes:

        1. `ResetZone` (optional);
        2. A sequence of `RequestRoute` tests for each route;
        3. A sequence of `WaitForRoute` tests for each (checks the state of
        the route and whether it is set in the right time);
        4. A sequence of `SetDetectorOccupied` and `SetDetectorFree` tests
        for each route (the latter are applied in conjuction with the route
        state ckecker tests `CheckIfRouteOccupied` or `CheckIfRouteFree`).

        Args:
            routes (list[Route]): A list of Route objects
            to generate the scenario for.
            add_zone_reset (bool): If True, adds a Zone reset test sequence
            at the beginning of the scenario.
            truncate (bool): If True, the passage scenario ends with
            the occupation of the route's furthest-from-its-startgate detector,
            which is also a trigger detector for some other route in
            the scenario. Otherwise, the scenario continues to the release of
            the route's last detector.
            fastmode (bool): If True, the time delays between the detector
            occupation and release events in the passage test sequence are
            ignored.

        Returns:
            SequentialMultiTest: The generated passage scenario wrapped into
            a SequentialMultiTest.
        """
        max_duration: float = 0
        tests: list[TestABC] = []
        # add the Reset test (optional)
        if add_zone_reset:
            zone: Zone | None = None
            for route in routes:
                if route.zone is not None:
                    zone = route.zone
                    break
            if zone is None:
                raise ValueError('Routes have no Zone assigned.')
            reset_zone_test = self._generate_reset_zone_test(zone)
            tests.append(reset_zone_test)
            max_duration += self._calculate_total_test_max_duration(
                reset_zone_test
                )
        # add the RouteRequest test for every route in the list
        route_request_tests = [
            self._generate_request_route_test(r) for r in routes
            ]
        tests.append(
            SequentialMultiTest(
                *route_request_tests,
                name=f'Request routes {", ".join(r.name for r in routes)}'
                )
            )
        for test in route_request_tests:
            max_duration += self._calculate_total_test_max_duration(test)
        # add a passage test sequence for each route
        route_passage_tests: list[SequentialMultiTest] = []
        route_passage_order = self.route_analyzer.determine_route_order(
            routes, truncate=truncate
            )
        max_duration_offset: float = 0
        for item in route_passage_order:
            passage_test = self._generate_passage_test_sequence(
                route=item.route,
                fastmode=fastmode,
                trigger=item.trigger_detector,
                last_detector_to_release=item.last_detector_to_release,
                check_elements_before_passage=False,
                wait_for_route_max_duration_offset=max_duration_offset
                )
            max_duration_offset += self._calculate_total_test_max_duration(
                passage_test
                )
            route_passage_tests.append(passage_test)
        tests.append(
            ParallelMultiTest(
                *route_passage_tests,
                name=(
                    f'Routes {", ".join(route.name for route in routes)} '
                    f'simultaneous passage'
                )
            )
        )
        max_duration += max_duration_offset
        # combine all the generated tests into a multipassage test scenario
        return SequentialMultiTest(
            *tests,
            max_duration=max_duration,
            name=(
                f'Routes {", ".join(route.name for route in routes)} '
                f'simultaneous passage simulation'
            )
        )

    def generate_passage_scenario(
        self,
        *routes: str | Route,
        add_zone_reset: bool = True,
        truncate: bool = True,
        fastmode: bool = True,
    ) -> SequentialMultiTest:
        """
        Generate a passage scenario for the provided route or a set of routes.

        The scenario includes:

        1. A `ResetZone` test (optional);
        2. A sequence of `RequestRoute` tests for each route;
        3. A sequence of `WaitForRoute` tests for each route (check the state
        of the route and its elements, and whether the route is set
        in the right time);
        4. A sequence of `SetDetectorOccupied` and `SetDetectorFree` tests
        for each route (the latter are applied in conjuction with the route
        state ckecker tests `CheckIfRouteOccupied` or `CheckIfRouteFree`).

        Args:
            routes (str | Route): A set of routes to generate the scenario for.
            Can be specified as route names (strings) or Route objects.
            add_zone_reset (bool): If True, adds the `ResetZone` test sequence
            at the beginning of the scenario. Defaults to True.
            truncate (bool): If True, the passage scenario ends with the
            occupation of the route's furthest-from-startgate detector,
            which is also a trigger detector for some other route in the
            scenario. (A detector is considered a trigger for a route if
            this detector must be released before this route may be passed.
            Such a detector usually releases the last common layout element
            shared between the this route and some route that has started
            being passed before it.) In case there are no such detectors,
            the fisrt detector in the layout is occupied. Without trucnation,
            the scenario continues to the release of the route's last detector.
            fastmode (bool): If True, the time delays between the detector
            occupation and release events in the passage test sequence are
            ignored. Defaults to True.

        Returns:
            SequentialMultiTest | None: The generated passage scenario wrapped
            into a SequentialMultiTest. If the route list is empty,
            the function returns None.
        """
        route_objs = self.route_manager._get_route_batch(routes)
        if not route_objs:
            raise ValueError('No routes have been found.')
        if len(route_objs) == 1:
            logger.info(
                'Generating a test scenario for the route %s passage',
                route_objs[0].name
                )
            return self._generate_passage_scenario_single_route(
                route_objs[0], add_zone_reset, truncate, fastmode
                )
        else:
            logger.info(
                'Generating a test scenario for the routes %s passage',
                ", ".join(r.name for r in route_objs)
                )
            return self._generate_passage_scenario_multiple_routes(
                route_objs, add_zone_reset, truncate, fastmode
                )



from ._detector import (
    DetectorSetFreeTest,
    DetectorSetOccupiedTest,
    DetectorResetTest
)


__all__ = [
    'DetectorSetFreeTest',
    'DetectorSetOccupiedTest',
    'DetectorResetTest'
]
